# Author: Edvinas Repečka IFM-0/2

from tkinter import Tk
from tkinter.filedialog import askdirectory, askopenfilename
import ecdsa
import numpy as np
import matplotlib.pyplot as plt
from colorama import init
import itertools

init()

class bcolors:
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

runApp = True
inputReq = ''

hd = 'ECCDS python app'
hd_1 = 'Please input required command:'

p_1 = '\t1 - Load private key'
p_2 = '\t2 - Load public key'
p_3 = '\t3 - Generate new ECC private and public keys'
p_4 = '\t4 - Export private and public keys'
p_5 = '\t5 - Load data file'
p_6 = '\t6 - Sign loaded file'
p_7 = '\t7 - Export signature'
p_8 = '\t8 - Load signature'
p_9 = '\t9 - Verify signature'
p_10 = '\t10 - Draw secp256k1 graph in real numbers'
p_11 = '\t11 - Export private key'
p_12 = '\t12 - Export public key'
p_13 = '\t13 - Draw secp256k1 graph over finite field'

nl = '\n'

p_e = '\texit/e - Exit app'

err_0 = 'Error: Unknown command'
err_1 = 'Error: Load or generate private ECC key first!'
err_2 = 'Error: Load or generate public ECC key first!'
err_3 = 'Error: Load file first!'
err_4 = 'Error: Load or generate signautre first!'

err_sign = 'Error: Invalid Digital Signature!'

i_c_1 = False
i_c_2 = False
i_c_3 = False
i_c_4 = False

c_1 = 'ECC private key loaded/generated'
c_2 = 'ECC public key loaded/generated'
c_3 = 'File data loaded'
c_4 = 'Signature loaded/generated'

w_1 = 'File loaded'
w_2 = 'ECC key generated'
w_3 = 'ECC key loaded'
w_4 = 'Signature loaded'

PrK = ''
PuK = ''

fileContentBytes = ''

Signature = ''

PrK_h = '--------------------------PrK--------------------------\n'
PuK_h = '--------------------------PuK--------------------------\n'

mainTxt = 'ECC python app\nInput requested command\n\n\n\n4-Export private and public keys\n\n\n7-'

def requestInput():    
    inputReq = input("Input command: ") 
    digestInput(inputReq)

def printError(err):
    print(bcolors.FAIL + err + bcolors.ENDC)

def printWarning(warning):
    print(bcolors.WARNING + warning + bcolors.ENDC)

def printOk(ok):
    print(bcolors.OKGREEN + ok + bcolors.ENDC)

def printHeader():
    if (i_c_1):
        printWarning(c_1)
    if (i_c_2):
        printWarning(c_2)
    if (i_c_3):
        printWarning(c_3)
    if (i_c_4):
        printWarning(c_4)
    print(hd + nl + hd_1 + nl + p_1 + nl + p_2 + nl+ p_3 + nl + p_4 + nl+ p_5 + nl + p_6 + nl+ p_7 + nl + p_8 + nl+ p_9 + nl + p_10 + nl+ p_11 + nl+ p_12 + nl + p_13 + nl + p_e)

def killApp():
    global runApp
    runApp = False

def exportSignature():
    if (i_c_4):
        dir = selectDirectory()
        if (not dir == ''):
            f = open(dir + "/App_Signature.txt", "w")
            f.write(str(Signature))
            f.close()
    else:
        if (not i_c_4):
            printError(err_4)

def exportPuK():
    if (i_c_2):
        dir = selectDirectory()
        if (not dir == ''):
            f = open(dir + "/App_PuK.txt", "w")
            f.write(PuK)
            f.close()
    else:
        if (not i_c_2):
            printError(err_2)

def exportPrK():
    if (i_c_1):
        dir = selectDirectory()
        if (not dir == ''):
            f = open(dir + "/App_PrK.txt", "w")
            f.write(PrK)
            f.close()
    else:
        if (not i_c_1):
            printError(err_1)

def exportBothKeys():
    if (i_c_1 and i_c_2):
        dir = selectDirectory()
        if (not dir == ''):
            f = open(dir + "/App_PrK.txt", "w")
            f.write(PrK)
            f.close()
            f = open(dir + "/App_PuK.txt", "w")
            f.write(PuK)
            f.close()
    else:
        if (not i_c_1):
            printError(err_1)
        if (not i_c_2):
            printError(err_2)

def selectDirectory():
    Tk().withdraw() 
    dir = askdirectory()
    return dir

def loadFile():
    Tk().withdraw() 
    filename = askopenfilename()
    s = ''
    if (not filename == ''):
        s = open(filename, 'r').read()
    return s

def generateECCKeys():
    global i_c_1, i_c_2
    global PrK, PuK

    eccKeys = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1)
    private_key = eccKeys.to_string().hex()
    puV = eccKeys.get_verifying_key()
    public_key = puV.to_string().hex()
    #public_key = base64.b64encode(bytes.fromhex(public_key))

    PrK = private_key
    PuK = public_key

    i_c_1 = True
    i_c_2 = True


def loadPublicKey():
    global PuK, i_c_2
    f_key = loadFile() 
    if (not f_key == ''):
        PuK = f_key
        i_c_2 = True

def loadPrivateKey():
    global PrK, i_c_1
    f_key = loadFile()  
    if (not f_key == ''):  
        PrK = f_key
        i_c_1 = True

def loadSignature():
    global Signature, i_c_4
    f_Sign = loadFile()   
    if (not f_Sign == ''): 
        Signature = f_Sign
        i_c_4 = True

def loadBinaryFile():
    Tk().withdraw() 
    filename = askopenfilename()
    s = ''
    if (not filename == ''):
        s = open(filename, mode='rb').read()
    return s

def loadTextFile():
    global fileContentBytes, i_c_3
    f_C = loadBinaryFile()
    if (not f_C == ''): 
        fileContentBytes = f_C
        i_c_3 = True

def signFile():
    if (i_c_1 and i_c_3):
        global i_c_4, Signature
        sk = ecdsa.SigningKey.from_string(bytes.fromhex(PrK), curve=ecdsa.SECP256k1)
        sign = sk.sign(fileContentBytes)
        Signature = sign.hex()
        i_c_4 = True
    else:
        if (not i_c_1):
            printError(err_1)
        if (not i_c_3):
            printError(err_3)

def verifySignature():
    if (i_c_2 and i_c_3 and i_c_4):
        vk = ecdsa.VerifyingKey.from_string(bytes.fromhex(PuK), curve=ecdsa.SECP256k1)
        try:
            verification = vk.verify(bytes.fromhex(Signature), fileContentBytes)
            if (verification):
                printOk("Signature verified.")
        except ecdsa.BadSignatureError:
            printError(err_sign)
        
    else:
        if (not i_c_2):
            printError(err_2)
        if (not i_c_3):
            printError(err_3)
        if (not i_c_4):
            printError(err_4)

def drawGraph():
    a = 0
    b = 7
    y, x = np.ogrid[-5:5:100j, -5:5:100j]
    plt.contour(x.ravel(), y.ravel(), pow(y, 2) - pow(x, 3) - x * a - b, [0])
    plt.grid()
    plt.show()

def isPrime(val):
    if val > 1:
        for i in range(2, int(val/2)+1):
            if (val % i) == 0:
                return False
        else:
            return True    
    else:
        return False

def drawGraphFiniteFields():
    p = input("Finite field = ")
    try:
        p = int(p)
    except ValueError:
        printError('Wrong input')
        return
    if (not isPrime(p)):
        printError(str(p) + ' is not a prime number')
        return
    rangeFrom = 0
    rangeTo = p
    a = 0
    b = 7
    # y^2 = x^3 + 7 mod p
    x,y = getEllipticCurvePoints(a,b,p)
    
    #np.ogrid[-5:5:100j, -5:5:100j]
    plt.scatter(x, y)
    plt.grid()
    plt.show()

def hasKey(dict, key):
    if (key in dict):
        return True
    else:
        return False

def l_f_x(x, a, b, p, x_l_f):
   x_l_f.setdefault((pow(x,3)+a*x+b)%p, []).append(x)

def l_f_y(y, p, y_l_f):
   y_l_f.setdefault((pow(y,2))%p, []).append(y)

def getEllipticCurvePoints(a,b,p):
    x_l_f = dict()
    y_l_f = dict()
    x = []
    y = []
    for num in range(0,p):
        l_f_x(num,a,b,p, x_l_f)
        l_f_y(num,p, y_l_f)

    allPoints = []
    for item in x_l_f.keys():
        if (hasKey(y_l_f, item)):
            tmp = list(itertools.product(x_l_f[item], y_l_f[item]))
            allPoints = allPoints + tmp
    allPoints = sorted(allPoints)
    for point in allPoints:
        x.append(point[0])
        y.append(point[1])
    return x,y

def digestInput(x):
    if (x == 'exit' or x == 'e'):
        killApp()
    elif (x == '1'):
        loadPrivateKey()
    elif (x == '2'):
        loadPublicKey()
    elif (x == '3'):
        generateECCKeys()
    elif (x == '4'):
        exportBothKeys()
    elif (x == '5'):
        loadTextFile()
    elif (x == '6'):
        signFile()
    elif (x == '7'):
        exportSignature()
    elif (x == '8'):
        loadSignature()
    elif (x == '9'):
        verifySignature()
    elif (x == '10'):
        drawGraph()
    elif (x == '11'):
        exportPrK()
    elif (x == '12'):
        exportPuK()
    elif (x == '13'):
        drawGraphFiniteFields()
    else:
        printError(err_0)

while runApp:
    printHeader()
    requestInput()
